import os
import sys
import subprocess
import time
import argparse

try:
    from rich.console import Console, Group
    from rich.table import Table
    from rich.live import Live
    from rich.panel import Panel
    from rich.spinner import Spinner
    from rich.align import Align
except ImportError:
    raise ImportError("Please install the rich package: pip install rich")


class PartionNames:
    GPU = "GPU"
    GPU_DEPINFO = "GPU-DEPINFO"


def ssh_remote_command(command):
    if args.proxy_host != "null":
        ssh_command = f"ssh -o StrictHostKeyChecking=no -J {username}@{proxy_host} {username}@{target_host} '{command}'"
    else:
        ssh_command = (
            f"ssh -o StrictHostKeyChecking=no {username}@{target_host} '{command}'"
        )
    try:
        result = subprocess.run(
            ssh_command,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.stdout.decode().strip()
    except subprocess.CalledProcessError as e:
        console.log(e.stderr.decode().strip())
        return e.stderr.decode().strip()


def parse_gpu_spec(spec):
    specs = {}
    if "gres:" in spec:
        spec = spec.split("gres:")[1]

    if "," in spec:
        for s in spec.split(","):
            if len(s.split(":")) == 3:
                _, gpu_model, n = s.split(":")
            else:
                gpu_model, n = s.split(":")

            specs[gpu_model] = int(n)

    else:
        if len(spec.split(":")) == 3:
            _, gpu_model, n = spec.split(":")
        else:
            gpu_model, n = spec.split(":")

        specs[gpu_model] = int(n)

    return specs


def req_dead_nodes():
    stdout = ssh_remote_command("sinfo -R -h")
    return [group for group in stdout.split() if "calcul" in group]


def req_gpu_used(nodes):
    for node in nodes:
        node["available_gpus"] = sum(node["specs"].values())

    for partition in PARTITIONS_TO_CHECK:
        stdout = ssh_remote_command(f'squeue -h -o "%20R %50b" -p "{partition}" -t r')
        groups = stdout.split()
        for nname, nspec in zip(groups[::2], groups[1::2]):
            # bash job
            if nspec == "N/A":
                continue

            for node in nodes:
                if node["name"] == nname:
                    parsed_nspecs = parse_gpu_spec(nspec)
                    for _, n in parsed_nspecs.items():
                        node["available_gpus"] -= n

    return nodes


def req_gpu_specs():
    deads = req_dead_nodes()
    nodes = []
    for partition in PARTITIONS_TO_CHECK:
        stdout = ssh_remote_command(
            command=f'sinfo -h -p "{partition}" -N -O "NodeList:30,Gres:50"'
        )
        groups = stdout.split()
        for node, spec in zip(groups[::2], groups[1::2]):
            nodes.append(
                {
                    "name": node,
                    "specs": parse_gpu_spec(spec),
                    "status": "dead" if node in deads else "alive",
                }
            )

    return req_gpu_used(nodes)


def update_table(table):
    live.update(
        Align.center(
            Panel(
                Group(spinner_fetching, table),
                title="GPU Status on LabHC Cluster",
            ),
            vertical="middle",
        )
    )
    nodes = req_gpu_specs()
    # fetching time
    for idx, node in enumerate(nodes):
        name = node["name"]
        n_gpus = sum(node["specs"].values())
        n_available_gpus = node["available_gpus"]
        gpu_models = "\n".join(
            [
                f"{n} * {gpu_model_mapping[gpu_model]['name']} ({gpu_model_mapping[gpu_model]['vram']}MB)"
                for gpu_model, n in node["specs"].items()
            ]
        )

        row_values = (
            (
                name,
                f"[bold green]{n_available_gpus}[/]/{n_gpus}"
                if n_available_gpus > 0
                else f"[bold red]{n_available_gpus}[/]/{n_gpus}",
                gpu_models,
            )
            if node["status"] == "alive"
            else (
                name,
                f"[bold red]down[/]",
                gpu_models,
            )
        )

        if len(table.rows) > idx:
            for row_idx, value in enumerate(row_values):
                table.columns[row_idx]._cells[idx] = value
        else:
            table.add_row(*row_values)

    last_updated = time.strftime("%H:%M:%S")

    # idle time
    for remaining in range(wait_every, 0, -1):
        spinner_idle = Spinner(
            "flip",
            text=f"[bold red]STATUS:[/] [yellow]Idle[/] ({remaining}s remaining) - Last updated at {last_updated}",
            style="bold yellow",
        )
        live.update(
            Align.center(
                Panel(
                    Group(spinner_idle, table),
                    title="GPU Status on LabHC Cluster",
                ),
                vertical="middle",
            )
        )
        time.sleep(1)

    return table


def init_table():
    table = Table(
        show_header=True,
        header_style="bold green",
        expand=True,
        show_edge=True,
        show_lines=True,
        min_width=70,
    )

    table.add_column("Node", style="bold dim italic", no_wrap=True)
    table.add_column("Availability", justify="center")
    table.add_column("GPU Models", justify="center")
    return table


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Check GPU status on LabHC cluster. Usage: python main.py <username>",
    )

    parser.add_argument(
        "username",
        type=str,
        help="Username on the cluster (required)",
    )

    parser.add_argument(
        "--wait_every", type=int, help="Wait time between checks", default=10
    )
    parser.add_argument(
        "--proxy_host",
        type=str,
        help="Proxy host",
        default="pass-lahc.univ-st-etienne.fr",
    )
    parser.add_argument(
        "--target_host", type=str, help="Target host", default="calcul-slurm-lahc-2"
    )
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    gpu_model_mapping = {
        "teslaa": {"name": "NVIDIA A40", "vram": 46068},
        "gtxp": {"name": "NVIDIA GeForce GTX 1080 Ti", "vram": 11264},
        "titanxm": {"name": "NVIDIA GeForce GTX TITAN X", "vram": 12288},
        "titanxp": {"name": "NVIDIA TITAN Xp", "vram": 12288},
        "titanxt": {"name": "NVIDIA TITAN RTX", "vram": 24576},
    }

    username = args.username
    proxy_host = args.proxy_host
    target_host = args.target_host
    wait_every = args.wait_every

    PARTITIONS_TO_CHECK = [PartionNames.GPU, PartionNames.GPU_DEPINFO]

    console = Console()
    table = init_table()
    spinner_fetching = Spinner(
        "dots",
        text="[bold red]STATUS:[/] [bold green]Fetch GPU Status[/]",
        style="green",
    )

    panel = Align.center(
        Panel(
            Group(spinner_fetching, table),
            title="GPU Status on LabHC Cluster",
        ),
        vertical="middle",
    )

    # clear the output
    os.system("clear")

    with Live(panel, refresh_per_second=3, console=console, screen=False) as live:
        while True:
            update_table(table)
            os.system("clear")
